/*
 * botnetop.c -- part of botnetop.mod
 *  
 * $Id: botnetop.c,v 1.3 2000/08/25 20:19:17 sup Exp $
 */
/*
 * Copyright (C) 2000  Teemu Hjelt <temex@iki.fi>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#define MODULE_NAME "botnetop"
#define MAKING_BOTNETOP
#define MODULE_VERSION "0.5.0"
#include "../module.h"
#include "../irc.mod/irc.h"
#include "../server.mod/server.h"
#include "../channels.mod/channels.h"

#undef global
static Function *global = NULL, *irc_funcs = NULL, *server_funcs = NULL, *channels_funcs = NULL;

#include "botnetop.h"

static int strict_host = 1;
static int bop_delay = 20;
static int bop_maxreq = 3;
static int bop_modeop = 0;
static int bop_linkop = 1;
static int bop_osync = 0;
static int bop_addhost = 0;
static int bop_setneed = 0;
static int bop_cycle = 0;
static int bop_log = 2;
static char bop_cycle_msg[256] = "Please part the channel so the bots can cycle!";
static struct delay_t *start_delay = NULL;

#include "misc.c"
#include "botcmds.c"
#include "ircbinds.c"
#include "botbinds.c"
#include "needbinds.c"
#include "tclcmds.c"
#include "delay.c"

static int botnetop_expmem()
{
  int size = 0;
  struct delay_t *d = start_delay;

  while (d) {
    size += sizeof(struct delay_t);
    d = d->next;
  } 
  return size;
}

static void botnetop_report(int idx, int details)
{
  int dtot = 0;
  struct delay_t *d = start_delay;

  while (d) {
    dtot++;
    d = d->next;  
  }
  if (details) {
    dprintf(idx, "    Tcl commands: bop_reqop, bop_letmein\n");
    dprintf(idx, "    %d delay record%s using %d bytes\n", dtot, dtot == 1 ? "" : "s", botnetop_expmem());
  }
}

static tcl_ints botnetop_tcl_ints[] =
{
  {"strict-host", &strict_host, 0},
  {"bop_delay", &bop_delay, 0},
  {"bop_maxreq", &bop_maxreq, 0},
  {"bop_modeop", &bop_modeop, 0},
  {"bop_linkop", &bop_linkop, 0},
  {"bop_osync", &bop_osync, 0},
  {"bop_addhost", &bop_addhost, 0},
  {"bop_setneed", &bop_setneed, 0},
  {"bop_cycle", &bop_cycle, 0},
  {"bop_log", &bop_log, 0},
  {0, 0, 0}
};

static tcl_strings botnetop_tcl_strings[] =
{
  {"bop_cycle_msg", bop_cycle_msg, 256, 0},
  {0, 0, 0, 0}
};

static char *botnetop_close()
{
  Context;
  del_hook(HOOK_SECONDLY, (Function) check_delays);
  rem_builtins(H_join, botnetop_join);
  rem_builtins(H_mode, botnetop_mode);
  rem_builtins(H_link, botnetop_link);
  rem_builtins(H_bot, botnetop_bot);
  rem_builtins(H_need, botnetop_need);
  rem_tcl_ints(botnetop_tcl_ints);
  rem_tcl_strings(botnetop_tcl_strings);
  rem_tcl_commands(botnetop_tcl_cmds);
  delay_free_mem();
  module_undepend(MODULE_NAME);
  return NULL;
}

char *botnetop_start();

static Function botnetop_table[] =
{
  (Function) botnetop_start,
  (Function) botnetop_close,
  (Function) botnetop_expmem,
  (Function) botnetop_report,
};

char *botnetop_start(Function * global_funcs)
{
  global = global_funcs;
  Context;
  module_register(MODULE_NAME, botnetop_table, 0, 5);
  if (!(irc_funcs = module_depend(MODULE_NAME, "irc", 1, 0)))
    return "You need the irc module to use the botnetop module.";
  if (!(server_funcs = module_depend(MODULE_NAME, "server", 1, 0)))
    return "You need the server module to use the botnetop module.";
  if (!(channels_funcs = module_depend(MODULE_NAME, "channels", 1, 0)))
    return "You need the channels module to use the botnetop module.";
  if (!module_depend(MODULE_NAME, "void", 305, 4))
    return "This module requires VoiD3.5.4 or later";
  add_hook(HOOK_SECONDLY, (Function) check_delays);
  add_builtins(H_join, botnetop_join);
  add_builtins(H_mode, botnetop_mode);
  add_builtins(H_link, botnetop_link);
  add_builtins(H_bot, botnetop_bot);
  add_builtins(H_need, botnetop_need);
  add_tcl_ints(botnetop_tcl_ints);
  add_tcl_strings(botnetop_tcl_strings);
  add_tcl_commands(botnetop_tcl_cmds);
  putlog(LOG_MISC, "*", "botnetop.mod v%s by Sup loaded.", MODULE_VERSION);
  return NULL;
}

