/*
 * delay.c -- part of botnetop.mod
 *
 * $Id$
 */
/* 
 * Copyright (C) 2000  Teemu Hjelt <temex@iki.fi>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

static struct delay_t *find_delay(char *chname)
{
  struct delay_t *delay = start_delay;

  while (delay) {
    if (!rfc_casecmp(delay->chan, chname))
      return delay;
    delay = delay->next;
  }
  return NULL;
}

static struct delay_t *add_delay(char *chname, char *bot)
{
  struct delay_t *newdelay;

  newdelay = (struct delay_t *) nmalloc(sizeof(struct delay_t));
  newdelay->next = start_delay;
  strncpy(newdelay->chan, chname, sizeof(newdelay->chan) - 1);
  strncpy(newdelay->bot, bot, sizeof(newdelay->bot) - 1);
  newdelay->reqtime = 0;
  newdelay->resptime = 0;
  newdelay->reqs = 0;
  start_delay = newdelay;
  putlog(LOG_DEBUG, "*", "botnetop.mod: New delay record created for %s", chname);
  return newdelay;
}

static void delay_free_mem()
{
  struct delay_t *d = start_delay, *dnext;

  while (d) {
    dnext = d->next;
    nfree(d);
    d = dnext;
  }
  start_delay = NULL;
}

static void check_delays()
{
  struct delay_t *d = start_delay;

  while (d) {
    if ((d->reqtime) && (d->reqtime <= now)) {
      d->reqtime = 0;
      bnop_askbot(d->bot, d->chan);
    }
    if ((d->resptime) && (d->resptime <= now)) {
      d->resptime = 0;
      d->reqs = 0;
    }
    d = d->next;
  }
}

