/* 
 * msgcmds.c -- part of irc.mod
 *   all commands entered via /MSG
 * 
 * $Id: msgcmds.c,v 1.15 2000/08/06 14:51:38 fabian Exp $
 */
/* 
 * Copyright (C) 1997  Robey Pointer
 * Copyright (C) 1999, 2000  Eggheads
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

static int msg_hello(char *nick, char *h, struct userrec *u, char *p)
{
  char host[161], s[161], s1[161];
  char *p1;
  int common = 0;
  int atr = 0;
  struct chanset_t *chan;
  struct flag_record fr = {FR_GLOBAL, 0, 0, 0, 0, 0};

  if ((!learn_users) && (!make_userfile))
    return 0;
  if (match_my_nick(nick))
    return 1;
  if (u)
    atr = u->flags;
  if (u && !(atr & USER_COMMON)) {
    dprintf(DP_HELP, "NOTICE %s :%s, %s.\n", nick, IRC_HI, u->handle);
    return 1;
  }
  if (get_user_by_handle(userlist, nick)) {
    struct flag_record fr = {FR_GLOBAL, 0, 0, 0, 0, 0};
    fr.global = atr;

    dprintf(DP_HELP, IRC_BADHOST1, IRC_BADHOST1_ARGS);
    dprintf(DP_HELP, IRC_BADHOST2, IRC_BADHOST2_ARGS);
    return 1;
  }
  simple_sprintf(s, "%s!%s", nick, h);
  if (u_match_mask(global_bans, s)) {
    dprintf(DP_HELP, "NOTICE %s :%s.\n", nick, IRC_BANNED2);
    return 1;
  }
  if (strlen(nick) > HANDLEN) {
    /* Crazy dalnet */
    dprintf(DP_HELP, "NOTICE %s :%s.\n", nick, IRC_NICKTOOLONG);
    return 1;
  }
  if (atr & USER_COMMON) {
    maskhost(s, host);
    strcpy(s, host);
    simple_sprintf(host, "%s!%s", nick, s + 2);
    userlist = adduser(userlist, nick, host, "-", USER_DEFAULT);
    putlog(LOG_MISC, "*", "%s %s (%s) -- %s",
	   IRC_INTRODUCED, nick, host, IRC_COMMONSITE);
    common = 1;
  } else {
    maskhost(s, host);
    if (make_userfile) {
      userlist = adduser(userlist, nick, host, "-",
		 sanity_check(default_flags | USER_MASTER | USER_OWNER));
      set_user(&USERENTRY_HOSTS, get_user_by_handle(userlist, nick),
	       "telnet!*@*");
    } else
      userlist = adduser(userlist, nick, host, "-",
			 sanity_check(default_flags));
    putlog(LOG_MISC, "*", "%s %s (%s)", IRC_INTRODUCED, nick, host);
  }
  for (chan = chanset; chan; chan = chan->next) {
    if (ismember(chan, nick))
      add_chanrec_by_handle(userlist, nick, chan->dname);
  }
  dprintf(DP_HELP, IRC_SALUT1, IRC_SALUT1_ARGS);
  dprintf(DP_HELP, IRC_SALUT2, IRC_SALUT2_ARGS);
  if (common) {
    dprintf(DP_HELP, "NOTICE %s :%s\n", nick, IRC_SALUT2A);
    dprintf(DP_HELP, "NOTICE %s :%s\n", nick, IRC_SALUT2B);
  }
  if (make_userfile) {
    fr.global = sanity_check(default_flags | USER_OWNER);

    dprintf(DP_HELP, "NOTICE %s :%s\n", nick, IRC_INITOWNER1);
    dprintf(DP_HELP, IRC_NEWBOT1, IRC_NEWBOT1_ARGS);
    dprintf(DP_HELP, IRC_NEWBOT2, IRC_NEWBOT2_ARGS);
    putlog(LOG_MISC, "*", IRC_INIT1, IRC_INIT1_ARGS);
    make_userfile = 0;
    write_userfile(-1);
    add_note(nick, origbotname, IRC_INITNOTE, -1, 0);
  } else {
    fr.global = default_flags;

    dprintf(DP_HELP, IRC_INTRO1, IRC_INTRO1_ARGS);
  }
  if (notify_new[0]) {
    sprintf(s, IRC_INITINTRO, nick, host);
    strcpy(s1, notify_new);
    while (s1[0]) {
      p1 = strchr(s1, ',');
      if (p1 != NULL) {
	*p1 = 0;
	p1++;
	rmspace(p1);
      }
      rmspace(s1);
      add_note(s1, origbotname, s, -1, 0);
      if (p1 == NULL)
	s1[0] = 0;
      else
	strcpy(s1, p1);
    }
  }
  return 1;
}

static int msg_pass(char *nick, char *host, struct userrec *u, char *par)
{
  char *old, *new;

  if (match_my_nick(nick))
    return 1;
  if (!u)
    return 1;
  if (u->flags & (USER_BOT | USER_COMMON))
    return 1;
  if (!par[0]) {
    dprintf(DP_HELP, "NOTICE %s :%s\n", nick,
	    u_pass_match(u, "-") ? IRC_NOPASS : IRC_PASS);
    putlog(LOG_CMDS, "*", "(%s!%s) !%s! PASS?", nick, host, u->handle);
    return 1;
  }
  old = newsplit(&par);
  if (!u_pass_match(u, "-") && !par[0]) {
    dprintf(DP_HELP, "NOTICE %s :%s\n", nick, IRC_EXISTPASS);
    return 1;
  }
  if (par[0]) {
    if (!u_pass_match(u, old)) {
      dprintf(DP_HELP, "NOTICE %s :%s\n", nick, IRC_FAILPASS);
      return 1;
    }
    new = newsplit(&par);
  } else {
    new = old;
  }
  putlog(LOG_CMDS, "*", "(%s!%s) !%s! PASS...", nick, host, u->handle);
  if (strlen(new) > 15)
    new[15] = 0;
  if (strlen(new) < 6) {
    dprintf(DP_HELP, "NOTICE %s :%s\n", nick, IRC_PASSFORMAT);
    return 0;
  }
  set_user(&USERENTRY_PASS, u, new);
  dprintf(DP_HELP, "NOTICE %s :%s '%s'.\n", nick,
	  new == old ? IRC_SETPASS : IRC_CHANGEPASS, new);
  return 1;
}

static int msg_op(char *nick, char *host, struct userrec *u, char *par)
{
  struct chanset_t *chan;
  char *pass;
  struct flag_record fr = {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};

  if (match_my_nick(nick))
    return 1;
  pass = newsplit(&par);
  if (u_pass_match(u, pass)) {
    /* Prevent people from gaining ops when no password set */
    if (!u_pass_match(u, "-")) {
      if (par[0]) {
	chan = findchan_by_dname(par);
	if (chan && channel_active(chan)) {
	  get_user_flagrec(u, &fr, par);
	  if (chan_op(fr) || (glob_op(fr) && !chan_deop(fr)))
	    add_mode(chan, '+', 'o', nick);
	    putlog(LOG_CMDS, "*", "(%s!%s) !%s! OP %s",
		   nick, host, u->handle, par);
	  return 1;
	}
      } else {
	chan = chanset;
	while (chan != NULL) {
	  get_user_flagrec(u, &fr, chan->dname);
	  if (chan_op(fr) || (glob_op(fr) && !chan_deop(fr)))
	    add_mode(chan, '+', 'o', nick);
	  chan = chan->next;
	}
	putlog(LOG_CMDS, "*", "(%s!%s) !%s! OP", nick, host, u->handle);
	return 1;
      }
    }
  }
  putlog(LOG_CMDS, "*", "(%s!%s) !*! failed OP", nick, host);
  return 1;
}

static int msg_key(char *nick, char *host, struct userrec *u, char *par)
{
  struct chanset_t *chan;
  char *pass;
  struct flag_record fr =
  {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};

  if (match_my_nick(nick))
    return 1;
  pass = newsplit(&par);
  if (u_pass_match(u, pass)) {
    /* Prevent people from getting key with no pass set */
    if (!u_pass_match(u, "-")) {
      if (!(chan = findchan_by_dname(par))) {
	dprintf(DP_HELP, "NOTICE %s :%s: /MSG %s key <pass> <channel>\n",
		nick, MISC_USAGE, botname);
	return 1;
      }
      if (!channel_active(chan)) {
	dprintf(DP_HELP, "NOTICE %s :%s: %s\n", nick, par, IRC_NOTONCHAN);
	return 1;
      }
      chan = findchan_by_dname(par);
      if (chan && channel_active(chan)) {
	get_user_flagrec(u, &fr, par);
	if (chan_op(fr) || (glob_op(fr) && !chan_deop(fr))) {
	  if (chan->channel.key[0]) {
	    dprintf(DP_SERVER, "NOTICE %s :%s: key is %s\n", nick, par,
		    chan->channel.key);
	    if (invite_key && (chan->channel.mode & CHANINV)) {
	      dprintf(DP_SERVER, "INVITE %s %s\n", nick, par);
	      putlog(LOG_CMDS, "*", "(%s!%s) !%s! KEY %s",
		     nick, host, u->handle, par);
	    }
	  } else {
	    dprintf(DP_HELP, "NOTICE %s :%s: no key set for this channel\n",
		    nick, par);
	    putlog(LOG_CMDS, "*", "(%s!%s) !%s! KEY %s", nick, host, u->handle,
		   par);
	  }
	}
	return 1;
      }
    }
  }
  putlog(LOG_CMDS, "*", "(%s!%s) !*! failed KEY", nick, host);
  return 1;
}

/* Don't have to specify a channel now and can use this command
 * regardless of +autovoice or being a chanop. (guppy 7Jan1999)
 */
static int msg_voice(char *nick, char *host, struct userrec *u, char *par)
{
  struct chanset_t *chan;
  char *pass;
  struct flag_record fr = {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};

  if (match_my_nick(nick))
    return 1;
  pass = newsplit(&par);
  if (u_pass_match(u, pass)) {
    if (!u_pass_match(u, "-")) {
      if (par[0]) {
	chan = findchan_by_dname(par);
	if (chan && channel_active(chan)) {
	  get_user_flagrec(u, &fr, par);
	  if (chan_voice(fr) || glob_voice(fr) ||
	      chan_op(fr) || glob_op(fr)) {
	    add_mode(chan, '+', 'v', nick);
	    putlog(LOG_CMDS, "*", "(%s!%s) !%s! VOICE %s",
		   nick, host, u->handle, par);
	  } else
	    putlog(LOG_CMDS, "*", "(%s!%s) !*! failed VOICE %s",
		nick, host, par);
	  return 1;
	}
      } else {
	chan = chanset;
	while (chan != NULL) {
	  get_user_flagrec(u, &fr, chan->dname);
	  if (chan_voice(fr) || glob_voice(fr) ||
	      chan_op(fr) || glob_op(fr))
	    add_mode(chan, '+', 'v', nick);
	  chan = chan->next;
	}
	putlog(LOG_CMDS, "*", "(%s!%s) !%s! VOICE", nick, host, u->handle);
	return 1;
      }
    }
  }
  putlog(LOG_CMDS, "*", "(%s!%s) !*! failed VOICE", nick, host);
  return 1;
}

static int msg_invite(char *nick, char *host, struct userrec *u, char *par)
{
  char *pass;
  struct chanset_t *chan;
  struct flag_record fr = {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};

  if (match_my_nick(nick))
    return 1;
  if (!u)
    return 0;
  pass = newsplit(&par);
  if (u_pass_match(u, pass)) {
    if (par[0] == '*') {
      for (chan = chanset; chan; chan = chan->next) {
	get_user_flagrec(u, &fr, chan->dname);
	if ((chan_op(fr) || (glob_op(fr) && !chan_deop(fr))) &&
	    (chan->channel.mode & CHANINV))
	  dprintf(DP_SERVER, "INVITE %s %s\n", nick, chan->name);
      }
      putlog(LOG_CMDS, "*", "(%s!%s) !%s! INVITE ALL", nick, host,
	     u->handle);
      return 1;
    }
    if (!(chan = findchan_by_dname(par))) {
      dprintf(DP_HELP, "NOTICE %s :%s: /MSG %s invite <pass> <channel>\n",
	      nick, MISC_USAGE, botname);
      return 1;
    }
    if (!channel_active(chan)) {
      dprintf(DP_HELP, "NOTICE %s :%s: %s\n", nick, par, IRC_NOTONCHAN);
      return 1;
    }
    /* We need to check access here also (dw 991002) */
    get_user_flagrec(u, &fr, par);
    if (chan_op(fr) || (glob_op(fr) && !chan_deop(fr))) {
      dprintf(DP_SERVER, "INVITE %s %s\n", nick, par);
      putlog(LOG_CMDS, "*", "(%s!%s) !%s! INVITE %s", nick, host,
	     u->handle, par);
      return 1;
    }
  }
  putlog(LOG_CMDS, "*", "(%s!%s) !%s! failed INVITE %s", nick, host,
	 u->handle, par);
  return 1;
}

static int msg_go(char *nick, char *host, struct userrec *u, char *par)
{
  struct chanset_t *chan;
  int ok = 0, ok2 = 0;
  struct flag_record fr = {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};

  if (match_my_nick(nick))
    return 1;
  if (!u)
    return 0;
  if (par[0]) {
    chan = findchan_by_dname(par);
    if (!chan)
      return 0;
    if (!(chan->status & CHAN_ACTIVE)) {
      putlog(LOG_CMDS, "*", "(%s!%s) !%s! failed GO (i'm blind)", nick, host,
	     u->handle);
      return 1;
    }
    get_user_flagrec(u, &fr, par);
    if (!chan_op(fr) && !(glob_op(fr) && !chan_deop(fr))) {
      putlog(LOG_CMDS, "*", "(%s!%s) !%s! failed GO (not op)", nick, host,
	     u->handle);
      return 1;
    }
    if (!me_op(chan)) {
      dprintf(DP_SERVER, "PART %s\n", chan->name);
      putlog(LOG_CMDS, chan->dname, "(%s!%s) !%s! GO %s", nick, host,
	     u->handle, par);
      return 1;
    }
    putlog(LOG_CMDS, chan->dname, "(%s!%s) !%s! failed GO %s (i'm chop)",
	   nick, host, u->handle, par);
    return 1;
  }
  chan = chanset;
  while (chan != NULL) {
    if (ismember(chan, nick)) {
      get_user_flagrec(u, &fr, par);
      if (chan_op(fr) || (glob_op(fr) && !chan_deop(fr))) {
	ok2 = 1;
	if (!me_op(chan)) {
	  dprintf(DP_SERVER, "PART %s\n", chan->name);
	  ok = 1;
	}
      }
    }
    chan = chan->next;
  }
  if (ok) {
    putlog(LOG_CMDS, "*", "(%s!%s) !%s! GO", nick, host, u->handle);
  } else if (ok2) {
    putlog(LOG_CMDS, "*", "(%s!%s) !%s! failed GO (i'm chop)", nick, host,
	   u->handle);
  } else {
    putlog(LOG_CMDS, "*", "(%s!%s) !%s! failed GO (not op)", nick, host,
	   u->handle);
  }
  return 1;
}

/* MSG COMMANDS
 *
 * Function call should be:
 *    int msg_cmd("handle","nick","user@host","params");
 *
 * The function is responsible for any logging. Return 1 if successful,
 * 0 if not.
 */
static cmd_t C_msg[] =
{
  {"go",		"",	(Function) msg_go,		NULL},
  {"mek-",		"",	(Function) msg_hello,		NULL},
  {"invite",		"o|o",	(Function) msg_invite,		NULL},
  {"key",		"o|o",	(Function) msg_key,		NULL},
  {"op",		"",	(Function) msg_op,		NULL},
  {"pass",		"",	(Function) msg_pass,		NULL},
  {"voice",		"",	(Function) msg_voice,		NULL},
  {NULL,		NULL,	NULL,				NULL}
};
